/* 
 * Copyright (c) 2011, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
/*
 *  ======== dman3_externalRM_initImpl.c ========
 *  Initialize the DMAN3 module
 */

/* This difine must precede inclusion of any xdc header files */
#define Registry_CURDESC ti_sdo_fc_dman3_desc

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/IHeap.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Registry.h>

#include <ti/sdo/fc/global/FCSettings.h>

#include "rmm.h"
#include "dman3.h"
#include "dman3_externalRM_impl.h"
#include <ti/sdo/fc/edma3/edma3_config.h>


#pragma CODE_SECTION(_DMAN3_initImpl, ".text:DMAN3_init");

__FAR__ xdc_runtime_IHeap_Handle _DMAN3_heapInt;
__FAR__ xdc_runtime_IHeap_Handle _DMAN3_heapExt;
__FAR__ xdc_runtime_IHeap_Handle _DMAN3_heapIdma3;

__FAR__ DMAN3_ScratchAllocFxn _DMAN3_allocScratch = NULL;
__FAR__ DMAN3_ScratchFreeFxn _DMAN3_freeScratch = NULL;

__FAR__ Uns _DMAN3_qdmaIndex = 0;

__FAR__ DMAN3_QdmaObj * DMAN3_QDMA;

EDMA3_RM_Handle _DMAN3_systemRMHandle = NULL;

UInt32          *_DMAN3_paRamAddr = NULL;
UInt32          *_DMAN3_edma3Addr = NULL;
UInt32          _DMAN3_nullPaRam = 0x0;

static Int _DMAN3_refCount = 0;

static Int regInit = 0;  /* Registry_addModule() called */

Registry_Desc ti_sdo_fc_dman3_desc;

/*
 *  ======== _DMAN3_initImpl ========
 *  Initialize the DMAN module
 */
Void _DMAN3_initImpl(Void)
{
    EDMA3_RM_Result rmStatus;
    EDMA3_RM_ResDesc resObj;
    Int i = 0;
    Registry_Result regResult;

    /*
     *  Register the module name for logging only once. Do this before any
     *  Log functions are called.
     */
    if (regInit == 0) {
        regResult = Registry_addModule(&ti_sdo_fc_dman3_desc, DMAN3_MODNAME);

        Assert_isTrue(regResult == Registry_SUCCESS, (Assert_Id)NULL);

        /* Set diags mask to FC default */
        FCSettings_init();
        FCSettings_setDiags(DMAN3_MODNAME);
        regInit = 1;
    }

    Log_print0(Diags_ENTRY, "[+E] _DMAN3_initImpl> Enter");

    if (_DMAN3_refCount++ == 0) {

        Assert_isTrue(NULL == _DMAN3_systemRMHandle, (Assert_Id)NULL);
        _DMAN3_systemRMHandle = EDMA3_getResourceManager(NULL, -1);
        Assert_isTrue(NULL != _DMAN3_systemRMHandle, (Assert_Id)NULL);

        /*
         *  Initialize _DMAN3_heapInt and _DMAN3_heapExt with provided
         *  heaps. Use internal heap for _DMAN3_heapInt when provided.
         *  (Actually this is a fatal error if ACPY3 uses internal DMA
         *  (IDMA) to write to PARAMs.
         */
        _DMAN3_heapInt = DMAN3_PARAMS.heapInternal;
        _DMAN3_heapExt = DMAN3_PARAMS.heapExternal;

        _DMAN3_heapInt = (_DMAN3_heapInt == NULL) ?
            _DMAN3_heapExt : _DMAN3_heapInt;
        _DMAN3_heapExt = (_DMAN3_heapExt == NULL) ?
            _DMAN3_heapInt : _DMAN3_heapExt;
        _DMAN3_heapIdma3 = (DMAN3_PARAMS.idma3Internal) ?
            _DMAN3_heapInt : _DMAN3_heapExt;

        /* This is a fatal configuration error. */
        Assert_isTrue(_DMAN3_heapInt != NULL, (Assert_Id)NULL);

        /*
         *  Configured function for allocating IDMA3 object's env from
         *  shared scratch memory.
         */
        _DMAN3_allocScratch = DMAN3_PARAMS.scratchAllocFxn;

        /*
         *  Configured function for allocating IDMA3 object's env from
         *  shared scratch memory.
         */
        _DMAN3_freeScratch = DMAN3_PARAMS.scratchFreeFxn;

        EDMA3_RM_getBaseAddress(_DMAN3_systemRMHandle, EDMA3_RM_CC_PHY_ADDR,
                (unsigned int *)(&_DMAN3_edma3Addr));
        Assert_isTrue(_DMAN3_edma3Addr != NULL, (Assert_Id)NULL);

        _DMAN3_paRamAddr = (Uns *)((Uns)_DMAN3_edma3Addr + PARAMOFFSET);

        /*
         *  If one of _DMAN3_allocScratch or _DMAN3_freeScratch is NULL,
         *  then the other must be NULL.
         */
        Assert_isTrue((_DMAN3_allocScratch != NULL) ||
                (_DMAN3_freeScratch == NULL), (Assert_Id)NULL);
        Assert_isTrue((_DMAN3_freeScratch != NULL) ||
                (_DMAN3_allocScratch == NULL), (Assert_Id)NULL);

        /* Acquire a PaRam that will be the nullPaRamIndex */
        resObj.type = EDMA3_RM_RES_PARAM_SET;
        resObj.resId = EDMA3_RM_RES_ANY;

        rmStatus = EDMA3_RM_allocResource(_DMAN3_systemRMHandle, &resObj);
        if (EDMA3_RM_SOK != rmStatus) {

            Log_print0(Diags_USER7, "[+7] _DMAN3_initImpl> "
                    "EDMA3 LLD-RM could not allocate the NULL PaRam Index."
                    " Resource not available !");

            Assert_isTrue(0, (Assert_Id)NULL);
        }
        _DMAN3_nullPaRam = resObj.resId;

        /*  GUNQ:
         *  Should I add the same code here that makes static, all QDMA
         *  channels mapped to this paRam ?
         */
        DMAN3_QDMA = (DMAN3_QdmaObj *)Memory_calloc(_DMAN3_heapExt,
                sizeof(DMAN3_QdmaObj) * DMAN3_PARAMS.numQdmaChannels,
                sizeof(Int), NULL);

        resObj.type = EDMA3_RM_RES_QDMA_CHANNEL;
        resObj.resId = EDMA3_RM_RES_ANY;

        for (i = 0; i < DMAN3_PARAMS.numQdmaChannels; i++) {
                        resObj.resId = EDMA3_RM_RES_ANY;
            rmStatus = EDMA3_RM_allocResource(_DMAN3_systemRMHandle, &resObj);

            if (EDMA3_RM_SOK != rmStatus) {

                Log_print0(Diags_USER7, "[+7] _DMAN3_initImpl> "
                        "EDMA3 LLD-RM could not allocate QDMA channels."
                        " Resource not available !");

                Assert_isTrue(0, (Assert_Id)NULL);
            }

            DMAN3_QDMA[i].index = resObj.resId;
            DMAN3_QDMA[i].available = TRUE;

            /* Value of 0 implies this can be used as scratch or persistent
               Value of >=1 implies it is being used as scratch */
            DMAN3_QDMA[i].scratch = 0;
        }
    }

    Log_print0(Diags_EXIT, "[+X] _DMAN3_initImpl> Exit");
}

/*
 *  ======== DMAN3_exit ========
 *  Finalization method of the DMAN3 module
 */
Void DMAN3_exit(Void)
{
    Log_print0(Diags_ENTRY, "[+E] DMAN3_exit> Enter");

    if (--_DMAN3_refCount == 0) {

        /* Release PaRam acquired  persistently */
        /* Not required since handle itself will be released */

        /* Release resource manager handle, check status */
        EDMA3_releaseResourceManager(NULL, -1);
                _DMAN3_systemRMHandle = NULL;
    }

    Log_print0(Diags_EXIT, "[+X] DMAN3_exit> Exit");
}
/*
 *  @(#) ti.sdo.fc.dman3; 1, 0, 4,403; 3-1-2011 18:56:44; /db/atree/library/trees/fc/fc-n29x/src/ xlibrary

 */

